//
//  Mutex.swift
//  Pix
//
//  Created by Jakob on 15.9.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Foundation
import os.log

/**
Wrapper of `NSCondition` providing debug logging.
*/
public class Mutex
{
	/// the debug level to use
	public enum DebugLevel
	{
		case
		/// no debug logging
		NONE,
		
		/// log name of current thread
		THREAD,
		
		/// log name of current thread and stack trace
		STACK
	}
	
	/// determines the debug level to use
	public static var debugLevel = DebugLevel.NONE

	public var name: String
	{
		return condition.name!
	}

	public init(_ name: String){
		Mutex.SYNC_QUEUE.sync {
			Mutex.number += 1
			condition.name = name + " \(Mutex.number)"
		}
		log("created")
	}

	public func lock()
	{
		log("waiting to acquire")
		condition.lock()
		log("acquired")
	}

	public func wait()
	{
		log("waiting for")
		condition.wait()
		log("finished waiting for")
	}

	public func wait(until: Date)
	{
		log("waiting for")
		condition.wait(until: until)
		log("finished waiting for")
	}

	public func signal()
	{
		log("signaling")
		condition.signal()
		log("signaled")
	}

	public func unlock()
	{
		condition.unlock()
		log("relinquished")
	}

	private static var number = 0
	private static let SYNC_QUEUE = DispatchQueue(label: "jakob.util.Mutex")

	private let condition = NSCondition()

	private func log(_ action: String)
	{
		if Mutex.debugLevel != .NONE //&& condition.name!.hasPrefix("video finished")
		{
			let thread = Thread.current
			var threadName = thread.name

			if threadName == nil {
				threadName = "<no name>"
			} else if threadName == "" && thread === Thread.main {
				threadName = "main"
			}
			if Mutex.debugLevel == .THREAD
			{
				os_log(
					"%@ mutex \"%@\" in thread \"%@\"",
					log: .default,
					type: .debug,
					action,
					condition.name!,
					threadName!
				)
			}
			else
			{
				var stackString = ""

				for s in Thread.callStackSymbols
				{
					stackString += "\n" + s
				}

				os_log(
					"%@ mutex \"%@\" in thread \"%@\", call stack is %@",
					log: .default,
					type: .debug,
					action,
					condition.name!,
					threadName!,
					stackString
				)
			}
		}
	}
}
