//
//  Updater.swift
//  PixUpdater
//
//  Created by Jakob on 26.1.19.
//  Copyright © 2019 Jakob. All rights reserved.
//

import Cocoa
import os.log

public class Updater
{
	private static let APP = "app"
	private static let updaterApp = "SoftwareUpdater"
	
	private static func executable(_ bundle: URL) -> URL
	{
		return
			bundle
				.appendingPathComponent("Contents")
				.appendingPathComponent("MacOS")
				.appendingPathComponent(updaterApp)
	}

	/**
	- parameters:
		- appName: e.g. "Pix"
		- wwwPath: the folder under www.jakob-spies.de where the app is stored, e.g. "pix"
	- throws: `Error` if Application Support/Pix unavailable
	*/
	public init(appName: String, wwwPath: String, isInFolder: Bool) throws
	{
		self.appName = appName
		urlPath = wwwPath
		self.isInFolder = isInFolder
		
		var appSupp =
			try FileManager.default.url(
				for: .applicationSupportDirectory,
				in: .userDomainMask,
				appropriateFor: nil,
				create: true
			)
		
		appSupp = appSupp.appendingPathComponent(appName)
		appSuppApp = appSupp.appendingPathComponent("update")
	}

	/**
	Do not call if version check is disabled in preferences!
	 - parameter excludeHigherOrEqualTo: disable update if version in WWW is ≥ `excludeHigherOrEqualTo`
	- throws: `Error`
	*/
	public func updateIfNecessary(excludeHigherOrEqualTo: [Int]? = nil) throws
	{
		DispatchQueue(label: "Jakob.Updater").async {
			do {
				let didStartUpdate = try self.checkVersion(excludeHigherOrEqualTo: excludeHigherOrEqualTo)
				
				if !didStartUpdate {
					self.cleanAppSupp()
				}
			} catch {
				Exception.alertUser(error, inf: "error in automatic version check/update")
			}
		}
	}


	//MARK: Private

	private let appName: String
	private let appSuppApp: URL
	private var updaterSuccessfullyStarted = false
	private let urlPath: String
	private let isInFolder: Bool

	private func cleanAppSupp()
	{
		let files: FileManager = FileManager.default

		if files.fileExists(atPath: appSuppApp.path)
		{
			do {
				try files.trashItem(at: appSuppApp, resultingItemURL: nil)
			} catch {
				Exception.alertUser(error, inf: "Cannot delete \(appSuppApp.path).")
			}
		}
	}

	/**
	Do not call if version check is disabled in preferences!
	*/
	private func checkVersion(excludeHigherOrEqualTo: [Int]? = nil) throws -> Bool
	{
		var didStartUpdate = false
		
		let vc = VersionCheck1(path: urlPath, name: appName, file: appName + ".zip")

		guard let myVersion = vc.localVersion else {
			Exception.alertUser("Cannot determine own version. Please report this as a bug!")
			return didStartUpdate
		}

		guard let wwwVersion: String = vc.internetVersion else {
			Exception.alertUser("Cannot determine version at \(VersionCheck1.SITE).")
			return didStartUpdate
		}
		
		if VersionCheck1.isObsolete(
			localVersion: myVersion,
			internetVersion: wwwVersion,
			excludeHigherOrEqualTo: excludeHigherOrEqualTo
		)
		{
			var doUpdate = false

			DispatchQueue.main.sync() {
				let alert = NSAlert()

				alert.alertStyle = .informational
				alert.addButton(withTitle: "Yes")
				alert.addButton(withTitle: "No")
				alert.messageText = "Automatic Update"


				alert.informativeText =
					"""
					\(appName) has version \(myVersion), but a newer version, \(wwwVersion), is available.
					Do you want to automatically update to the new version?
					"""
				doUpdate = alert.runModal() == .alertFirstButtonReturn
			}

			if doUpdate
			{
				do {
					try update()
					didStartUpdate = true
				} catch let e as Exception {
					e.display()
				} catch {
					Exception.alertUser(error)
				}
			}
		}
		return didStartUpdate
	}

	/**
	- throws:`Exception` and `Error`
	*/
	private func update() throws
	{
		let myBundle: Bundle = Bundle.main
		let files = FileManager.default

		guard let guts = myBundle.url(forResource: Updater.updaterApp, withExtension: Updater.APP) else {
			throw Exception(msg: "Cannot find \(Updater.updaterApp).app in my own guts. Please report this as a bug!")
		}
		
		guard let bundleId = myBundle.bundleIdentifier else {
			throw Exception(msg: "Cannot find own bundle ID. Please report this as a bug!")
		}

		let gutsExec = Updater.executable(guts).path
		
		if !files.isExecutableFile(atPath: gutsExec) {
			throw Exception(msg: "\"\(gutsExec)\" is not executable")
		}

		var isDir: ObjCBool = false
		let appSuppAppPath: String = appSuppApp.path
		var appSuppAppExists = files.fileExists(atPath: appSuppAppPath, isDirectory: &isDir)
		
		os_log("[Updater.update] appSuppApp \"%{public}@\" exists: %d", log: .default, type: .info, appSuppAppPath, appSuppAppExists)

		if appSuppAppExists {
			if isDir.boolValue
			{
				os_log("[Updater.update] trashing contents of \"%{public}@\"", log: .default, type: .info, appSuppAppPath)
			
				for item in try files.contentsOfDirectory(at: appSuppApp, includingPropertiesForKeys: nil) {
					try files.trashItem(at: item, resultingItemURL: nil)
				}
			} else {
				try files.removeItem(at: appSuppApp)
				appSuppAppExists = false
			}
		}

		if !appSuppAppExists {
			os_log("[Updater.update] trying to create \"%{public}@\"", log: .default, type: .info, appSuppAppPath)
			try files.createDirectory(at: appSuppApp, withIntermediateDirectories: true, attributes: nil)
		}

		let updAppUrl = appSuppApp.appendingPathComponent(Updater.updaterApp + "." + Updater.APP)
//		let encoder = JSONEncoder()
//		let parmFile = appSuppApp.appendingPathComponent(SoftwareUpdateParameters.fileName)
//
//		let parm =
//			SoftwareUpdateParameters(
//				bundleId: bundleId,
//				shortAppName: appName,
//				urlPath: urlPath,
//				isInFolder: isInFolder
//			)
//
//		let data: Data = try encoder.encode(parm)

		try files.copyItem(at: guts, to: updAppUrl)

		os_log(
			"[Updater.update] copied \"%{public}@\" to \"%{public}@\"",
			type: .info,
			guts.path,
			updAppUrl.path
		)

//		try data.write(to: parmFile)
				
		let updater = Process()
		let execUrl = Updater.executable(updAppUrl)
		let execPath: String = execUrl.path
		
		if !files.fileExists(atPath: execPath) {
			throw Exception(msg: "\"\(execPath)\" does not exist")
		}

		//		try QuarantineHandler.deleteQuarantineAttr(execUrl)

		if !files.isExecutableFile(atPath: execPath) {
			throw Exception(msg: "\"\(execPath)\" is not executable")
		}
		
		updater.executableURL = execUrl
		updater.currentDirectoryURL = appSuppApp
		updater.arguments = [bundleId, appName, urlPath, String(isInFolder)]
		
		os_log(
			"[Updater.update] trying to start \"%{public}@\" in directory \"%{public}@\" with parameters %{public}@",
			type: .info,
			execPath,
			appSuppAppPath,
			updater.arguments!
		)
		
		try updater.run()
		os_log("[Updater.update] started \"%{public}@\"", log: .default, type: .info, execUrl.path)
		updaterSuccessfullyStarted = true
	}
}
