//
//  VersionCheck.swift
//  Pix
//
//  Created by Jakob on 1.4.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Cocoa
import os.log

/// Version checker.
public class VersionCheck1
{
	//MARK: Public

	public static let SITE = "https://www.jakob-spies.de/"

	/**
	Compares local version and version on `SITE`.
	- returns: whether the version in the Internet is greater than the local version. If one of the version strings is not
		properly interpretable, it returns `false`.
	*/
	public static func isObsolete(
		localVersion: String,
		internetVersion: String,
		excludeHigherOrEqualTo: [Int]? = nil
	)
		-> Bool
	{
		guard let locTokens = intVersion(localVersion) else {
			os_log("Local version %{public}@ contains non-integer component.", type: .error, localVersion)
			return false
		}

		guard let wwwTokens = intVersion(internetVersion) else {
			os_log("Internet version %{public}@ contains non-integer component.", type: .error, internetVersion)
			return false
		}

		guard
				excludeHigherOrEqualTo == nil
			|| wwwTokens.lexicographicallyPrecedes(excludeHigherOrEqualTo!)
		else {
			return false
		}

		return locTokens.lexicographicallyPrecedes(wwwTokens)
	}

	public var internetVersion: String?
	{
		if cachedInternetVersion == nil
		{
			getInternetVersion()

			guard let data = data else {
				return cachedInternetVersion
			}

			guard let text = String(bytes: data, encoding: .utf8) else {
				os_log("cannot read %{public}@", log: OSLog.default, type: .error, address)
				return cachedInternetVersion
			}

			guard let range = text.range(of: VersionCheck1.H3_START) else	{
				os_log("cannot find \"%{public}@\" in %@", log: OSLog.default, type: .error, VersionCheck1.H3_START, address)
				return cachedInternetVersion
			}

			let sub = text.suffix(from: range.upperBound)

			guard let h3 = sub.range(of: VersionCheck1.H3_END) else {
				os_log("cannot find %{public}@ where it should be", log: OSLog.default, type: .error, VersionCheck1.H3_END)
				return cachedInternetVersion
			}

			cachedInternetVersion = String(sub.prefix(upTo: h3.lowerBound))
			os_log("found version %{public}@ at %@", log: OSLog.default, type: .info, cachedInternetVersion!, root)
		}

		return cachedInternetVersion
	}

	public var localVersion: String?
	{
		return Bundle.main.infoDictionary?["CFBundleShortVersionString"] as? String
	}

	/**
	Constructor.
	- parameters:
		- path: "pix"
		- name: "Pix"
		- file: "Pix.zip"
	*/
	public init(path: String, name aName: String, file aFile: String){
		root = VersionCheck1.SITE + path + "/"
		address = root + "index.html"
		name = aName
		file = aFile
	}

	/**
		Version check.
	*/
	public func checkVersion()
	{
		taskQueue.async {
			self.doCheckVersion()
		}
	}

	/// Non-private for testing.
	func checkVersion(_ myVersion: String)
	{
		if let publicVersion = internetVersion
		{
			if VersionCheck1.isObsolete(localVersion: myVersion, internetVersion: publicVersion)
			{
				DispatchQueue.main.async {
					let alert = NSAlert()

					alert.informativeText = "\(self.name) version is \(myVersion), but \(self.root)\(self.file) is already at version \(publicVersion)."
					alert.messageText  = "Automatic Version Check"
					alert.alertStyle = .warning
					alert.runModal()
				}
			}
		}
	}

	//MARK: Private Static

	private static let H3_START = ">New in Version "
	private static let H3_END = "</h"

	private static func version(_ text: String) -> [String]
	{
		return text.components(separatedBy: ".")
	}
	
	private static func intify(_ tok: [String]) -> [Int]?
	{
		var res: [Int]?
		
		do {
			res = try tok.map() {
				guard let v = Int($0) else {
					throw Exception(msg: "")
				}
				return v
			}
		} catch {
			// NOP
		}
		
		return res
	}
	
	private static func intVersion(_ versionId: String) -> [Int]?
	{
		return intify(version(versionId))
	}

	//MARK: Private Instance Members

	private let join = NSCondition()
	private var data: Data?
	private var myError: Error?
	private let root: String
	private let address: String
	private let taskQueue = DispatchQueue(label: "VersionCheck")
	private let name: String
	private let file: String
	private var cachedInternetVersion: String?

	private func doCheckVersion()
	{
		if let myVersion = localVersion
		{
			checkVersion(myVersion)
		}
		else {
			os_log("cannot obtain own version", log: OSLog.default, type: .error)
		}
	}

	private func getInternetVersion()
	{
		var urlRequest = URLRequest(url: URL(string: address)!)

		urlRequest.setValue("no-cache", forHTTPHeaderField: "Cache-Control")

		let getTask = URLSession.shared.dataTask(with: urlRequest) {
				downloaded, response, error in

				if let error = error {
					self.myError = error
				} else {
					self.data = downloaded
				}
				self.join.lock()
				self.join.signal()
				self.join.unlock()
			}

		join.lock()
		getTask.resume()

		let timeLimit = Date(timeIntervalSinceNow: 30)

		while myError == nil && data == nil {
			join.wait(until: timeLimit)
		}
		join.unlock()

		if let err = myError {
			os_log("cannot download %{public}@ because of %{public}@", log: OSLog.default, type: .error, address, err as CVarArg)
		} else if data == nil {
			os_log("cannot download %{public}@", log: OSLog.default, type: .error, address)
		}
	}
}
