//
//  DownloadTask.swift
//  Filmbeschaffer
//
//  Created by Jakob on 27.10.20.
//  Copyright © 2020 Jakob. All rights reserved.
//

import Foundation

/**
Abstract base class of the `URLSessionTask` wrappers.
*/
class DownloadTask
{
	static let semaphore = VariableSemaphore(Preferences.instance.maxConnections)

	/// the wrapped `URLSessionTask`
	var sessionTask: URLSessionTask
	{
		fatalError("sessionTask not implemented in abstract DownloadTask")
	}
	
	/**
	Tries to get a connection and to resume the `sessionTask`.
	- parameter timeout: maximal time to wait for a connection
	- returns: whether resume was executed
	*/
	func resume(_ timeout: TimeInterval) -> Bool
	{
		let res = DownloadTask.semaphore.acquire(timeout: timeout)
		
		if res {
			sessionTask.resume()
		}
		
		return res
	}

	/**
	Tries to get a connection and to `resume` the wrapped `URLSessionTask`.
	Returns immediately and executes the
	`completionHandler` when it is clear whether a connection could be obtained in time.
	- parameters:
		- timeout: maximal time to wait for a connection
		- completionHandler: what to do (in addition to resuming the wrapped `URLSessionTask`)
			if a connection was obtained (argument: `true`) or not (argument: `false`)
	*/
	func resume(_ timeout: TimeInterval, completionHandler: @escaping (Bool) -> Void)
	{
		DownloadTask.semaphore.acquire(timeout: timeout) { haveConn in
			if haveConn {
				self.sessionTask.resume()
			}
			completionHandler(haveConn)
		}
	}

	/**
	Cancels the `sessionTask` and releases the connection.
	*/
	func cancel() {
		sessionTask.cancel()
		release()
	}
	
	/**
	Cancels the `sessionTask` and releases the connection. Call this when the
	`DownloadTask` has finished on its own.
	*/
	func release() {
		DownloadTask.semaphore.release()
	}
}
