//
//  VariableSemaphore.swift
//  Filmbeschaffer
//
//  Created by Jakob on 25.10.20.
//  Copyright © 2020 Jakob. All rights reserved.
//

import Foundation
import AppUtilLib
import os.log

/**
A semaphore the capacity of which can be changed.
It has an internal counter with values between 0 and the capacity.
*/
class VariableSemaphore
{
	//MARK: Private

	/**
	The maximal value of the counter. Must be positive.
	*/
	var capacity: Int
	{
		set {
			if newValue <= 0 {
				fatalError("cannot set capacity to non-positive value \(newValue)")
			}
	
			let diff = newValue - _capacity
			
			if diff != 0 {
				serialQueue.sync {
					let wouldBeCount = count + diff

					// `_capacity - count` items are checked out.
					// `hangover` is the excess of this number over the new capacity.
					hangover = max(0, -wouldBeCount) // = _capacity - count - newValue

					_capacity = newValue
					count = max(0, wouldBeCount)
					
					if count == diff { // <--> `count` was 0, but is now > 0
						notify()
					}
				}
			}
		}
		
		get {
			_capacity
		}
	}
	
	/**
	Construct with given capacity.
	
	_Precondition:_ `size > 0`
	- parameter size: the capacity
	*/
	init(_ size: Int) {
		guard size > 0 else {
			fatalError("cannot set capacity to non-positive value \(size)")
		}
		_capacity = size
		count = size
	}
	
	/**
	Waits until the counter is positive or the end time is reached.
	Decreases the counter in the first of these cases.
	- parameter until: the maximal point in time until we may wait
		for the counter to become positive
	- returns: whether the counter became positive
	*/
	func acquire(until: Date) -> Bool
	{
		var acquired = false
		
		serialQueue.sync {
//			os_log("[VariableSemaphore.acquire] found count = %i", type: .debug, count)
			if count > 0 {
				count -= 1
				acquired = true
			} else {
				positive.lock()
			}
		}
		
		while !acquired {
//			os_log("[VariableSemaphore.acquire] have to wait", type: .debug, count)
			if Date() >= until {
				positive.unlock()
				break
			}
			
			positive.wait(until: until)
			positive.unlock()

			serialQueue.sync {
				if count > 0 {
					count -= 1
					acquired = true
				} else {
					positive.lock()
				}
			}
		}
//		os_log("[VariableSemaphore.acquire] returning %d", type: .debug, acquired)

		return acquired
	}

	/**
	Waits until the counter is positive or the timeout is reached.
	Decreases the counter in the first of these cases.
	- parameter timeout: the maximal time to wait for
		the counter to become positive
	- returns: whether the counter became positive
	*/
	func acquire(timeout: TimeInterval) -> Bool
	{
		return acquire(until: Date().addingTimeInterval(timeout))
	}
	
	/**
	Increases the counter if it is below the capacity.
	*/
	func release() {
//		os_log("[VariableSemaphore.release] entered", type: .debug)
		
		serialQueue.sync {
			if hangover > 0 {
				hangover -= 1
			} else if count < _capacity {
				count += 1
				
				if count == 1 { // i.e. it was 0, so there can be waiting threads
					notify()
				}
			}
		}
	}
	
	/**
	Waits (in another thread) until the counter is positive or the end time is reached,
	but returns immediately.
	Decreases the counter in the first of these cases.
	- parameters:
		- until: the maximal point in time until we may wait
			for the counter to become positive
		- completionHandler: what to do if the counter became 0 (argument: `true`)
			or not (argument: `false`)
	*/
	func acquire(until: Date, completionHandler: @escaping (Bool) -> Void)
	{
		concurrentQueue.async {
			completionHandler(self.acquire(until: until))
		}
	}
	
	/**
	Waits (in another thread) until the counter is positive or the timeout is reached,
	but returns immediately.
	Decreases the counter in the first of these cases.
	- parameters:
		- timeout: the maximal time to wait for the counter to become positive
		- completionHandler: what to do if the counter became 0 (argument: `true`)
			or not (argument: `false`)
	*/
	func acquire(timeout: TimeInterval, completionHandler: @escaping (Bool) -> Void)
	{
		let end = Date().addingTimeInterval(timeout)
		
		concurrentQueue.async {
			completionHandler(self.acquire(until: end))
		}
	}

	//MARK: Private
	
	/// Invariant: `0 ≤ count ≤ capacity`
	private var count: Int

	private var _capacity: Int
	
	/**
	If this is positive, then `release` reduces hangover instead of increasing counter.
	It gets positive when the capacity is reduced by more than `count`.
	
	_Invariant:_ `hangover >= 0`
	*/
	private var hangover = 0

	/// For synchronizing access to our internal state.
	private let serialQueue = DispatchQueue(label: "VariableSemaphore's serial queue")

	/// For executing tasks asynchronously and concurrently.
	private let concurrentQueue = DispatchQueue(label: "VariableSemaphore's concurrent queue", attributes: .concurrent)

	private let positive = Mutex("VariableSemaphore's Mutex")
	
	private func notify() {
		positive.lock()
		positive.signal()
		positive.unlock()
	}
}
