//
//  DownloadSkeleton.swift
//  Filmbeschaffer
//
//  Created by Jakob on 8.12.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Cocoa

/// Abstract implementation of `Download`.
class DownloadSkeleton: NSObject, Download, URLSessionDownloadDelegate
{
	struct Event
	{
		let time: Date
		let bytes: Int64
	}

	/**
	Use this global connection counter whenever you start, resume, suspend, or finish a download!
	*/
	static let connectionSemaphore = VariableSemaphore(Preferences.instance.maxConnections)
	
	static let waitForConn = 1.0 // s

	//MARK: public

	var errorMessage: String?
	{
		guard let err = downloader.theError else {
			return nil
		}

		let res: String

		if err is DownloadError {
			res = (err as! DownloadError).message
		} else {
			res = err.localizedDescription
		}

		return res
	}

	var url: URL
	{
		downloadUrl
	}
	
	var status: DownloadState
	{
		_isNewState = false
		return state
	}

	var downloaded: Int64
	{
		return fileDownloadedBytes
	}
	
	var totalSize: Int64
	{
		get { fatalError("totalSize not implemented in DownloadSkeleton") }
	}

	var isFinished: Bool
	{
		return state == .finished
	}

	var isNewState: Bool
	{
		return _isNewState
	}

	var isResumable: Bool
	{
		return state == .cancelled || downloader.haveError
	}

	var isActive: Bool
	{
		return state == .downloading || state == .preparing
	}

	var state = DownloadState.preparing
	{
		didSet {
			_isNewState = true
		}
	}

	var speed: Double
	{
		fatalError("speed not implemented in DownloadSkeleton")
	}


	init(url u: URL, target to: URL) throws
	{
		target = to
		downloadUrl = u
		minPast = Preferences.instance.speedCalculationTimeSpan
	}

	func cancel() {
		fatalError("cancel not implemented in DownloadSkeleton")
	}

	func start() {
		fatalError("start not implemented in DownloadSkeleton")
	}

	func destroy() {
		internalSynchronizationQueue.sync { // prevent other threads from changing fileDownloadTask
			guard let dlt = fileDownloadTask else {
				return
			}
			dlt.cancel()
			fileDownloadTask = nil
		}
	}

	func setError(_ err: Error?) {
		downloader.theError = err
	}

	func openVideo() {
		fatalError("openVideo not implemented in abstract class DownloadSkeleton")
	}

	//MARK: URLSessionDownloadDelegate
	
	func urlSession(
		_ session: URLSession,
		downloadTask: URLSessionDownloadTask,
		didFinishDownloadingTo location: URL)
	{
//		os_log("urlSession(didFinishDownloadingTo) entered", type: .debug)
		do {
			let files: FileManager = Downloader.FILES

			try files.moveItem(at: location, to: target)
			try files.setAttributes(Downloader.FILE_ATTR, ofItemAtPath: target.path)
			signalFinished(.finished)
		} catch {
			setError(error)
			signalFinished(.error)
		}
	}
	
	func urlSession(
		_ session: URLSession,
		downloadTask: URLSessionDownloadTask,
		didWriteData: Int64,
		totalBytesWritten: Int64,
		totalBytesExpectedToWrite: Int64
	){
			progress(bytes: totalBytesWritten, total: totalBytesExpectedToWrite)
	}

	func urlSession(_ session: URLSession, task: URLSessionTask, didCompleteWithError error: Error?)
	{
//		os_log("urlSession(didCompleteWithError) entered", type: .debug)
	
		// This method also gets called when the task is canceled,
		// which we do not regard as an error.
		if let err = error,
			state != .cancelled
		{
			setError(err)
			
			// This was already called in FileDownload.cancel.
			signalFinished(DownloadState.error)
		}
	}

	//MARK: protected
	
	var fileDownloadedBytes: Int64 = 0
	var fileDownloadTask: FileDownloadTask?
	var fileDownloadTotalSize: Int64 = 0

	/// from the DataDownloadTasks
	var downloadedBytes: Int64 = 0

	let downloadUrl: URL
	var target: URL
	let downloader = Downloader()
	let downloadQueue = DispatchQueue(label: "Download Queue")
	let internalSynchronizationQueue	= DispatchQueue(label: "Internal Synchronization Queue of Download")
	let minPast: TimeInterval

	/**
	How many bytes were downloaded at which point in time?
	 
	 Take care to synchronize access to this array!
	
	_Invariant:_ `!events.isEmpty`
	
	_Invariant:_ `events.map{ $0.time }` is strictly monotonously increasing
	*/
	var events = [Event](repeating: Event(time: Date(), bytes: 0), count: 1)
	
	/**
	To be called in response to a request to resume the `downloadTask`.
	- parameter haveConn: Could the download be started/resumed?
	*/
	func resumeWorked(_ haveConn: Bool) {
		if haveConn {
			state = .downloading
		} else {
			internalSynchronizationQueue.sync {
				if state == .waiting {
					fileDownloadTask?.resume(FileDownload.waitForConn, completionHandler: resumeWorked(_:))
				}
			}
		}
	}

	func signalFinished(_ newState: DownloadState)
	{
//		os_log("signalFinished entered with %@", type: .debug, "\(newState)" as CVarArg)
		
		internalSynchronizationQueue.sync {
//			if state != .cancelled {
				state = newState
				fileDownloadTask?.release()
//			}
			fileDownloadTask = nil
		}
	}
	
	func progress(bytes: Int64, total: Int64) {
		let now = Date()

		fileDownloadedBytes = bytes
		fileDownloadTotalSize = total

		internalSynchronizationQueue.sync { // otherwise crash possible because `events` is not thread-safe
			let n = events.count
			var i = 0
			
			while i < n && now.timeIntervalSince(events[i].time) >= minPast
			{
				i += 1
			}
			
			if i > 1  // stupid Swift cannot build empty range 0..<0
			{
				events.removeSubrange(0 ..< i - 1)
			}
			
			events.append(Event(time: now, bytes: downloadedBytes + fileDownloadedBytes))
		}
	}
	
	/**
	Starts the download, waits for a connection, but does not wait for the
	download to finish.
	Runs in `downloadQueue`.
	*/
	func download(_ videoUrl: URL?)
	{
		// prevent other threads from changing downloadTask
		internalSynchronizationQueue.sync
		{
			let data = downloader.resumeData
			
			if urlSession == nil {
				urlSession =
					URLSession(configuration: URLSessionConfiguration.default, delegate: self, delegateQueue: nil)
			}
			
			fileDownloadTask = data == nil ?
					FileDownloadTask(urlSession: urlSession!, url: videoUrl!)
				: 	FileDownloadTask(urlSession: urlSession!, resumeData: data!)
			
			state = .waiting
			fileDownloadTask!.resume(FileDownload.waitForConn, completionHandler: resumeWorked(_:))
		}
	}

	
	//MARK: private
	private var _isNewState = false
	private var urlSession: URLSession?
}
