//
//  HtmlUtil.swift
//  Filmbeschaffer
//
//  Created by Jakob on 31.5.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Foundation
import os.log

/**
 Utility functions for HTTP. Static.
 */
class HtmlUtil
{
	//MARK: Public
	
	static let URL_COMPONENT_SEPARATOR = "/"
	
	/**
	- throws: `DownloadError`
	*/
	static func find(_ token: String, in text: String, desc: Any) throws -> Range<String.Index>
	{
		guard let rg = text.range(of: token) else {
			throw DownloadError("\(desc) does not contain \(token)")
		}
		return rg
	}

	static func isHttpUrl(_ spec: String) -> Bool
	{
		spec.hasPrefix("http://") || spec.hasPrefix("https://")
	}

	/**
	- throws: `DownloadError`
	*/
	static func find(_ token: Character, in text: String.SubSequence, lim: String, desc: Any) throws -> String.Index
	{
		guard let p = text.firstIndex(of: token) else {
			throw DownloadError("no \(token) after \"(lim)\" in \(desc)")
		}
		return p
	}

	/**
	- throws: `DownloadError`
	*/
	static func toUrl(_ text: String) throws -> URL
	{
		guard let url = URL(string: text) else {
			throw DownloadError(text + " is not a URL")
		}
		return url
	}

	/**
	- throws: `DownloadError`
	*/
	static func urlAttribute(_ attr: String, in json: Any) throws -> URL
	{
		return try toUrl(stringAttribute(attr, in: json))
	}

	/**
	- throws: `Error`
	*/
	static func deserialize(_ html: String) throws -> Any
	{
//		os_log("trying to deserialize\n%@", log: OSLog.default, type: .debug, html as CVarArg)

		guard let data = html.data(using: .utf8) else {
			throw DownloadError("Cannot encode \"" + html + "\"")
		}
		do {
			return try JSONSerialization.jsonObject(with: data)
		}catch{
			os_log("cannot deserialize\n%@", log: OSLog.default, type: .error, html as CVarArg)
			throw error
		}
	}

	/**
	Tries to find and deserialize a JSON object within HTML code.

	- parameters:
		- token: the string marking the start of the JSON string, like an attribute name
		- in: the String containing the JSON string
		- endJson: the String marking the end of the JSON String
		- offset: the number of steps from the last character of the JSON String to the first character of `endJson`
		- desc: text describing the search area for the error message if the search fails

	- returns:  the result of the deserialization
	- throws: `DownloadError`, `Error`
	*/
	static func findJson(_ token: String, in text: String, endJson: String, offset: Int, desc: Any) throws -> Any
	{
		os_log("seeking JSON object %@", log: OSLog.default, type: .debug, token as CVarArg)

		var rg = try find(token, in: String(text), desc: desc)
		var html = text[rg.upperBound...]
		var p = try find("{", in: html, lim: token, desc: desc)
		let string = String(html[p...])

		rg = try find(endJson, in: string, desc: "\(desc) after \(token)")
		p = rg.lowerBound

		if string.index(string.startIndex, offsetBy: offset) > p
		{
			// Otherwise the program crashes on calculating string.index(p, offsetBy: -offset).
			throw DownloadError(endJson + " is closer than \(offset) to the start of " + string)
		}
		html = string[...string.index(p, offsetBy: -offset)]

		let x = try deserialize(String(html))

		os_log("found and deserialized JSON object %@", log: OSLog.default, type: .debug, token as CVarArg)
		return x
	}

	/**
	- throws: `DownloadError`
	*/
	static func attribute(_ attr: String, in json: Any, descNotDict: Any, descMissing: Any) throws -> Any
	{
		return try attribute(attr, in: toDictionary(json, desc: descNotDict), desc: descMissing)
	}

	/**
	- throws: `DownloadError`
	*/
	static func attribute(_ attr: String, in json: Any) throws -> Any
	{
		return try attribute(attr, in: json, descNotDict: json, descMissing: json)
	}

	/**
	- throws: `DownloadError`
	*/
	static func stringAttribute(_ attr: String, in json: Any) throws -> String
	{
		let value = try attribute(attr, in: json, descNotDict: json, descMissing: json)

		guard let a = value as? String else {
			throw DownloadError("attribute " + attr + " of \(json) is not a String")
		}
		return a
	}

	/**
	- throws: `DownloadError`
	*/
	static func firstObjOfArrayAttribute(_ attr: String, in json: Any, desc: Any? = nil) throws -> Any
	{
		guard let obj = try optFirstObjOfArrayAttribute(attr, in: json, desc: desc) else {
			throw DownloadError("attribute \(attr) is an empty array")
		}

		return obj
	}
	
	/**
	- throws: `DownloadError`
	*/
	static func optFirstObjOfArrayAttribute(_ attr: String, in json: Any, desc: Any? = nil) throws -> Any?
	{
		return try optArrayAttribute(attr, in: json, desc: desc)?.first
	}

	/**
	- throws: `DownloadError`
	*/
	static func arrayAttribute(_ attr: String, in json: Any, desc: Any? = nil) throws -> Array<Any>
	{
		let value = try attribute(attr, in: json)
		return try toArray(value, desc: desc ?? json)
	}

	/**
	- throws: `DownloadError`
	*/
	static func optArrayAttribute(_ attr: String, in json: Any, desc: Any? = nil) throws -> Array<Any>?
	{
		let value = try attribute(attr, in: json)
		return try value is NSNull ? nil : toArray(value, desc: desc ?? json)
	}

	/**
	- throws: `DownloadError`
	*/
	static func toDictionary(_ json: Any, desc: Any) throws -> [String: Any]
	{
		guard let dic = json as? [String: Any] else {
			throw DownloadError("\(desc) is not a JSON dictionary")
		}
		return dic
	}

	/**
	- throws: `DownloadError`
	*/
	static func toArray(_ json: Any, desc: Any) throws -> Array<Any>
	{
		guard let arr = json as? Array<Any> else {
			throw DownloadError("\(desc) is not a JSON array")
		}
		return arr
	}
	
	static func makeUrl(base: URL, spec: String) -> URL?
	{
		let singleUrl: URL?

		if isHttpUrl(spec)
		{
			singleUrl = URL(string: spec)
		}
		else if spec.hasPrefix(URL_COMPONENT_SEPARATOR)
		{
			// URL relative to root
			singleUrl = URL(string: base.scheme! + "://" + base.host! + spec)
		}
		else
		{
			// probably a relative URL

			var baseSpec = base.deletingLastPathComponent().absoluteString
						
			baseSpec = HtmlUtil.stripUrlParameters(baseSpec)
			
			if !baseSpec.hasSuffix(URL_COMPONENT_SEPARATOR) {
				baseSpec += URL_COMPONENT_SEPARATOR
			}
			singleUrl = URL(string: baseSpec + spec)
		}
		return singleUrl
	}
	
	static func makeRelative(_ uri: String, to: URL) throws -> String
	{
		var res = uri
		
		guard isHttpUrl(uri) else {
			return res
		}

		guard let u = URL(string: uri) else {
			throw DownloadError("strange URI \(uri)")
		}

		guard to.contains(u) else {
			return res
		}
		
		// This is of course incorrect in general, but it worked for all cases we encountered so far.
		// For a correct implementation, see optRelPath() in UrlUtil.swift:487 of Pix.
		res = u.lastPathComponent
		
		return res
	}

	static func stripUrlParameters(_ res: String) -> String
	{
		let pos = res.firstIndex(of: "?")

		return pos == nil ? res : String(res.prefix(upTo: pos!))
	}

	static func stripUrlParameters(_ u: URL) -> URL
	{
		return u.query == nil ? u : URL(string: stripUrlParameters(u.absoluteString))!
	}

	
	//MARK: Private

	/**
	- throws: `DownloadError`
	*/
	private static func attribute(_ attr: String, in dic: [String: Any], desc: Any) throws -> Any
	{
		guard let a = dic[attr] else {
			throw DownloadError("\(desc) has no attribute " + attr)
		}
		return a
	}
	
	/**
	 - returns: `u` with removed URL parameters and removed trailing slash, if it has a path
	 */
	fileprivate static func strip(_ u: URL) -> URL
	{
		var v = stripUrlParameters(u)
		let p = v.path
		
		if !p.isEmpty && p != "/" {
			let last = v.lastPathComponent
			v = v.deletingLastPathComponent().appendingPathComponent(last) // removes trailing '/'
		}
		return v
	}

//	/**
//	 Containment of stripped URLs.
//	 
//	 - parameters:
//			- parent: the potential parent, without URL parameters or trailing slash, if it has a path
//			- child: the potential parent, without URL parameters or trailing slash, if it has a path
//	 - returns: whether `child` is equal to or a descendant of `parent` in the forrest af all URLs
//	 */
//	fileprivate static func strippedContains(parent: URL, child: URL) -> Bool
//	{
//		return
//				parent == child
//			||
//					child.pathComponents.count > parent.pathComponents.count
//				&& strippedContains(parent: parent, child: child.deletingLastPathComponent())
//	}
}

extension URL
{
	/**
	- parameter other: a URL
	- returns: whether `other` is equal to or a descendant of `self` in the forrest af all URLs. Trailing slashes and URL parameters are ignored.
  */
	func contains(_ other: URL) -> Bool
	{
		// The stripping is necessary since a trailing '/' changes the URL, provided that the URL has a path.
		// Cf. e.g. https://ahrefs.com/blog/trailing-slash/#:~:text=A%20trailing%20slash%20is%20a,not%20have%20the%20trailing%20slash.
		return
				HtmlUtil.strip(self) == HtmlUtil.strip(other)
			||
					other.pathComponents.count > pathComponents.count
				&& contains(other.deletingLastPathComponent())
	}
}
