//
//  StreamDownload.swift
//  Filmbeschaffer
//
//  Created by Jakob on 8.12.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Cocoa
import AppUtilLib
import os.log

class StreamDownload: DownloadSkeleton
{
	//MARK: Static
	
	/**
	 Splits by `\n` and `\r\n`. Public for unit test.
	 */
	static func toLines(_ text: String) -> [Substring]
	{
		var result = [Substring]()
		let zero: String.Index = text.startIndex
		var index = zero
		var isBetweenLines = false
		let end = text.endIndex
		var lineStart: String.Index?
		
		while index < end {
			let ch = text[index]
			
			if isLinebreak(ch) {
				if !isBetweenLines {
					isBetweenLines = true
					
					if let start = lineStart {
						result.append(text[start ..< index])
					}
				}
			} else if isBetweenLines {
				lineStart = index
				isBetweenLines = false
			} else if lineStart == nil {
				lineStart = index
			}
			index = text.index(after: index)
		}
		
		if !isBetweenLines, let start = lineStart {
			result.append(text[start ..< index])
		}
		
		return result
	}
	
	private static let METADATA 				= "#"
	private static let EXT_X 					= METADATA + "EXT-X-"
	private static let STREAM_INF 			= EXT_X + "STREAM-INF"
	private static let MEDIA 					= EXT_X + "MEDIA"
	private static let I_FRAME_STREAM_INF	= EXT_X + "I-FRAME-STREAM-INF"
	private static let MAP						= EXT_X + "MAP"
	private static let BANDWIDTH 				= "BANDWIDTH="
	
	private static let NEW_LINE_STRING 		= "\n"
	private static let NEW_LINE 				= Character(NEW_LINE_STRING)
	private static let LINE_BREAK 			= NEW_LINE_STRING.data(using: .utf8)!
	
	/// Relative local location of the secondary M3U8 file if the HTTP URL given in the master M3U8 file is an absolute URL.
	private static let LOC_FOR_SEGMENTS		= "medias"
	
	private static func isDigit(_ c: Character) -> Bool
	{
		return CharacterSet.decimalDigits.contains(c.unicodeScalars.first!)
	}
	
	/// - throws: `DownloadError`
	private static func extractUri(_ line: String) throws -> Range<String.Index>
	{
		guard let rg = line.range(of: "URI=\"") else {
			throw DownloadError("no URI found in \(line)")
		}
		
		let pos1 = rg.upperBound
		
		guard let pos2 = line.suffix(from: pos1).firstIndex(of: "\"") else {
			throw DownloadError("URI specification  in '\(line)' does not end with '\"'.")
		}
		
		return pos1 ..< pos2
	}
	
	/// - throws: `DownloadError`
	private static func extractResolution(_ line: String) throws -> String
	{
		guard let rg = line.range(of: "RESOLUTION=") else {
			throw DownloadError("no RESOLUTION found in \(line)")
		}
		
		let pos1 = rg.upperBound
		let optPos2 = line.suffix(from: pos1).firstIndex(of: ",")
		
		return optPos2 == nil ? String(line.suffix(from: pos1)) : String(line[pos1 ..< optPos2!])
	}
	
	/**
	 - returns: Let _u_ be the the URI contained in the line.
	 If _u_ is a child of `base`, then the relative path is returned. Otherwise _u_ itself.
	 - throws: `DownloadError`
	 */
	private static func extractLocation(_ line: String, to base: URL) throws -> (String, Range<String.Index>)
	{
		let rg = try StreamDownload.extractUri(line)
		let str = String(line[rg])
		
		return try (HtmlUtil.makeRelative(str, to: base), rg)
	}

	private static func isLinebreak(_ ch: Character) -> Bool
	{
		return ch == "\n" || ch == "\r\n" || ch == "\r" // Would you have known that "\r\n" is one Character?
	}
	
	/**
	 Replaces the URI spec at `rg` in the `line` with its last path component appended to `LOC_FOR_SEGMENTS`.
	 - returns the resulting `String` and the URI spec
	 */
	private static func replaceLocation(_ line: String, rg: Range<String.Index>) throws -> String
	{
		let spec = String(line[rg])
		var res = line
		
		guard let url = URL(string: spec) else {
			throw DownloadError("bug in StreamDownload: \"\(spec)\" in \"\(line)\" is not a URL")
		}
		
		let loc: String = StreamDownload.LOC_FOR_SEGMENTS + "/" + url.lastPathComponent
		
		res.replaceSubrange(rg, with: loc)
		
		return res
	}

	private static func determineLocation(lines: inout [String], index i: Int, base: URL) throws -> String
	{
		let line = lines[i]
		var (loc, rg) = try StreamDownload.extractLocation(line, to: base)
		
		if HtmlUtil.isHttpUrl(loc) {
			// not a relative location
			lines[i] = try StreamDownload.replaceLocation(line, rg: rg)
		}
		return loc
	}

	
	//MARK: Public Instance Members
	
	override var totalSize: Int64
	{
		let tot: Int64
		
		if keepMaster {
			tot = fileDownloadTotalSize == 0 ? 0 : downloadedBytes + fileDownloadTotalSize
		} else {
			tot = Int64(round(Double(segmentUrl.count) * averageSegmentSize))
		}
		return tot
	}
	
	override var downloaded: Int64
	{
		return downloadedBytes + fileDownloadedBytes
	}
	
	override var speed: Double
	{
		let now = Date()
		
		let nBytes = downloaded
		// because the value could be already greater when we reach the end of the method
		
		var event: Event?
		var diff = 1.0
		
		internalSynchronizationQueue.sync {
			let n = events.count
			var i = n - 1
			
			while i >= 0
			{
				event = events[i]
				diff = now.timeIntervalSince(event!.time)
				
				if diff >= minPast {
					break
				}
				i -= 1
			}
		}
		
		return Double(nBytes - event!.bytes) / diff
	}
	
	override var url: URL
	{
		userUrl
	}
	
	init(url: URL, target: URL, isArte: Bool) throws
	{
		indexUrl = URL(fileURLWithPath: "index.m3u8", isDirectory: false, relativeTo: target)
		userUrl = url
		try super.init(url: isArte ? ArteUrlFinder().findVideoUrl(url) : url, target: target)
		state = .waiting
	}
	
	override func cancel() {
		state = .cancelling
		destroy()
		state = .cancelled
	}
	
	override func openVideo() {
		NSWorkspace.shared
			.open(
				[indexUrl],
				withAppBundleIdentifier: "Jakob.Pix",
				options: [],
				additionalEventParamDescriptor:  nil,
				launchIdentifiers: nil
			)
	}
	
	override func start()
	{
		downloadQueue.async {
			do {
				self.state = .preparing
				try self.download(self.downloadUrl, canBeMaster: true)
			} catch let e as DownloadError {
				DispatchQueue.main.async {
					let alert = NSAlert(error: e)
					
					alert.messageText = alert.messageText + "\n" + e.message
					alert.runModal()
				}
			} catch {
				DispatchQueue.main.async {
					NSAlert(error: error).runModal()
				}
			}
		}
	}
	
	override func destroy() {
		internalSynchronizationQueue.sync { // prevent other threads from changing downloadTasks
			for dlt in downloadTasks.values {
				dlt.cancel()
			}
			downloadTasks.removeAll()
		}
		super.destroy()
	}
	
	//MARK: Private Instance Members
	
	// Should be `volatile`, but this does not exist in Swift.
	private var downloadTasks = [URL : DataDownloadTask]()
	
	private var segmentUrl = Set<URL>() // avoid downloading the same file twice
	
	/// Should `processMaster()` write the current M3U8 file to disk?
	private var keepMaster = false
	
	private let indexUrl: URL
	private var locationOfSegmentsM3u8: String?
	private var locationOfSegments: URL?
	private var iFrames = [String]()
	
	// relative or absolute URLs of the "media"
	private var media = [String]()
	
	private var existingFolders = [String]()
	
	/// The original URL the user entered.
	private let userUrl: URL
	
	private var averageSegmentSize: Double
	{
		var avg = 3700000.0
		let numDownloadedSegments = events.count - 1
		
		if numDownloadedSegments != 0 {
			avg = Double(downloadedBytes) / Double(numDownloadedSegments)
		}
		//		os_log("average segment size after %u downloaded: %f", log: .default, type: .debug, numDownloadedSegments, avg)
		return avg
	}
	
	private func download(_ m3uUrl: URL, canBeMaster: Bool) throws
	{
		os_log("[download] entered with canBeMaster = %d and m3uUrl = %@", type: .debug, canBeMaster, m3uUrl as CVarArg)
		
		var headers = [String:String]()
		
		if !canBeMaster {
			headers["Referer"] = downloadUrl.absoluteString
		}
		
		let text = try downloader.getUtf8TextPage(m3uUrl, headers: headers)
		let lines = StreamDownload.toLines(text)
		let n: Int = lines.count
		
		downloadedBytes += Int64(downloader.data!.count)
		
		if canBeMaster {
			os_log("[download] text:\n%@\n#lines: %u", type: .debug, text, n)
		}
		
		if n < 2 {
			throw DownloadError("\(m3uUrl) contains less than 2 lines")
		}
		
		if lines[0] != "#EXTM3U" {
			throw DownloadError("\(m3uUrl) is not an M3U file, the first line is \"\(lines[0])\"")
		}
		
		var isMaster = canBeMaster
		
		if isMaster
		{
			isMaster = false
			
			for line in lines {
				if line.hasPrefix(StreamDownload.STREAM_INF) {
					isMaster = true
				}
				if line.hasPrefix(StreamDownload.MEDIA) {
					keepMaster = true
				}
				if isMaster && keepMaster {
					break
				}
			}
		}
		
		if isMaster {
			try processMaster(lines, masterM3uUrl: m3uUrl)
		} else {
			let base = HtmlUtil.stripUrlParameters(m3uUrl)
			try prepareSegmentsDownload(lines, base: base)
			try downloadSegments(base: base)
		}
	}
	
	/**
	 Finds the M3U URL for the segments, fills `iFrames` and `media`, and calls `download(`.`, canBeMaster: false)`.
	 
	 - parameters:
	 - url: URL of the master M3U8 file
	 - lines: the lines of the master M3U8 file
	 */
	private func processMaster(_ linesSub: [Substring], masterM3uUrl: URL) throws
	{
		var maxBandWidth = 0
		var bestM3u: String?
		var nextLineIsStreamInf = false
		var bestStreamInfLine: String?
		
		// modifiable copy
		var lines = linesSub.map() { String($0) }
		
		for line in lines
		{
			if nextLineIsStreamInf
			{
				bestM3u = line
				nextLineIsStreamInf = false
			}
			else if line.hasPrefix(StreamDownload.STREAM_INF)
			{
				//TBD take care to not catch "AVERAGE-BANDWIDTH", which is different from "BANDWIDTH"
				if let rg = line.range(of: StreamDownload.BANDWIDTH)
				{
					let endLine = line.index(before: line.endIndex)
					let equalsPos = rg.upperBound
					
					if equalsPos >= endLine {
						throw DownloadError("corrupt line \"\(line)\": nothing after \"\(StreamDownload.BANDWIDTH)\"")
					}
					
					let sub = line.suffix(from: equalsPos)
					let number = sub.prefix(while: { StreamDownload.isDigit($0) })
					
					if let bandwidth = Int(number) {
						if bandwidth > maxBandWidth {
							maxBandWidth = bandwidth
							bestStreamInfLine = line
							nextLineIsStreamInf = true
						}
					} else {
						throw DownloadError("corrupt line \"\(line)\"")
					}
				}
			}
		}
		
		guard let m3u = bestM3u else {
			throw DownloadError("corrupt master M3U, could not find best version")
		}
		//			os_log("bestm3u: %@", type: .debug, m3u as CVarArg)
		
		guard let singleUrl = HtmlUtil.makeUrl(base: masterM3uUrl, spec: m3u) else {
			throw DownloadError("corrupt master M3U: malformed URL \(m3u)")
		}
		
		let resolution = try StreamDownload.extractResolution(bestStreamInfLine!)
		
		if keepMaster {
			if let m3uUrl = URL(string: m3u) { // m3u describes an absolute HTTP URL
				locationOfSegmentsM3u8 = StreamDownload.LOC_FOR_SEGMENTS + "/" + m3uUrl.lastPathComponent
			} else {
				// we assume that m3u is the relative location of the M3U8 file with respect to the master M3U8
				locationOfSegmentsM3u8 = m3u
			}
		}
		
		var linesToDelete = Set<Int>()
		let base = masterM3uUrl.deletingLastPathComponent()
		
		for i in 0 ..< lines.count {
			let line = lines[i]
			
			if line.hasPrefix(StreamDownload.I_FRAME_STREAM_INF) {
				if try resolution == StreamDownload.extractResolution(line) {
					try iFrames.append(StreamDownload.determineLocation(lines: &lines, index: i, base: base))
				} else {
					linesToDelete.insert(i)
				}
			} else if line.hasPrefix(StreamDownload.MEDIA) {
				try media.append(StreamDownload.determineLocation(lines: &lines, index: i, base: base))
			}
		}
		
		if keepMaster {
			for i in 0 ..< lines.count {
				let line = lines[i]
				
				if line.hasPrefix(StreamDownload.STREAM_INF) && line != bestStreamInfLine {
					linesToDelete.insert(i)
					linesToDelete.insert(i+1)
				}
			}
			
			var text = Data(capacity: downloader.data!.count)
			
			let doReplace =
					locationOfSegmentsM3u8 != nil
				&& locationOfSegmentsM3u8!.hasPrefix(StreamDownload.LOC_FOR_SEGMENTS) //TBD isn't this automatically implied?
			
			for i in 0 ..< lines.count {
				if !linesToDelete.contains(i) {
					var line = String(lines[i])
					
					if doReplace && line == m3u {
						line = locationOfSegmentsM3u8!
					}
					text.append((line + StreamDownload.NEW_LINE_STRING).data(using: .utf8)!)
				}
			}
			
			try writeData(text, toFile: target.appendingPathComponent("index.m3u8"))
		}
		
		try downloadIFrames(base)
		try downloadMedia(base)
		try download(singleUrl, canBeMaster: false)
	}

	/**
	 - parameter baseUrl: parent URL of the master M3U8 file
	 - throws: whatever `downloader.throwOnError()` throws
	 */
	private func downloadIFrames(_ baseUrl: URL) throws
	{
		for path in iFrames {
			if state == .cancelling || state == .cancelled {
				break
			}

			let downloadUrl: URL
			let targetFile: URL
			
			if let url = URL(string: path) {
				targetFile =
					target
						.appendingPathComponent(StreamDownload.LOC_FOR_SEGMENTS)
						.appendingPathComponent(url.lastPathComponent)
				
				downloadUrl = url
			} else {
				downloadUrl = baseUrl.appendingPathComponent(path)
				targetFile = target.appendingPathComponent(path)
			}
				
			try startDownloadTask(url: downloadUrl, targetFile: targetFile)
		}
	}
	
	/**
	 - throws: `DownloadError`, `Error`
	 */
	private func downloadMedia(_ baseUrl: URL) throws
	{
		for pathOrUrl in media
		{
			if state == .cancelling || state == .cancelled {
				break
			}

			let mediaM3u8: URL
			let m3u8Location: URL

			if let url = URL(string: pathOrUrl) {
				m3u8Location =
					target
						.appendingPathComponent(StreamDownload.LOC_FOR_SEGMENTS)
						.appendingPathComponent(url.lastPathComponent)
				
				mediaM3u8 = url
			} else {
				mediaM3u8 = baseUrl.appendingPathComponent(pathOrUrl)
				m3u8Location = target.appendingPathComponent(pathOrUrl)
			}

			let text: String = try downloader.getUtf8TextPage(mediaM3u8)
			let lines = StreamDownload.toLines(text)
			var paths = Set<Substring>()
			let data: Data = downloader.data!
			
			downloadedBytes += Int64(data.count)
			try writeData(data, toFile: m3u8Location)
			
			for line in lines {
				if !line.hasPrefix(StreamDownload.METADATA) {
					paths.insert(line)
				}
			}
			
			for path in paths {
				try startDownloadTask(
					url: mediaM3u8.deletingLastPathComponent().appendingPathComponent(String(path)),
					targetFile: m3u8Location.deletingLastPathComponent().appendingPathComponent(String(path))
				)
			}
		}
	}
	
	/**
	 - throws: whatever `downloader.throwOnError()` throws
	 */
	private func startDownloadTask(url: URL, targetFile: URL) throws
	{
		let downloadTask =
			DataDownloadTask(url){ tempFile, response, error in
				self.downloadCompletionHandler(url: url, targetFile: targetFile, data: tempFile, response: response, err: error)
			}

		internalSynchronizationQueue.sync {
			downloadTasks[url] = downloadTask
		}

		os_log(
			"[startDownloadTask] starting download of %@ to %@",
			type: .debug,
			url as CVarArg,
			targetFile as CVarArg
		)

		while state != .cancelling && state != .cancelled && !downloadTask.resume(1) {}
	}

	/**
	 Fills `segmentUrl`, writes the M3U8 file for the segments, sets `locationOfSegments`.
	 */
	private func prepareSegmentsDownload(_ lines: [Substring], base: URL) throws
	{
		let comment = "\n# URL entered by user was \(userUrl)\n# M3U master location was \(downloadUrl)".data(using: .utf8)!
		var newData = Data(capacity: downloader.data!.count + comment.count)
		var isFirst = true

		os_log("[prepareSegmentsDownload] base: %@", type: .debug, base as CVarArg)

		for line in lines {
			if state == .cancelling {
				break
			}

			if line.hasPrefix(StreamDownload.METADATA)
			{
//				os_log("[prepareSegmentsDownload] line with metadata prefix: %@", type: .debug, line as CVarArg)
				if !isFirst {
					newData.append(StreamDownload.LINE_BREAK)
				}
		
				newData.append(line.data(using: .utf8)!)
				
				if isFirst {
					isFirst = false
					newData.append(comment)
				}
			}
			else if let tsUrl = HtmlUtil.makeUrl(base: base, spec: String(line))
			{
//				os_log("[prepareSegmentsDownload] line without metadata prefix: %@", type: .debug, line as CVarArg)
//				os_log("[prepareSegmentsDownload] tsUrl: %@", type: .debug, tsUrl as CVarArg)
				let fileName = HtmlUtil.stripUrlParameters(tsUrl.lastPathComponent)

				segmentUrl.insert(tsUrl)
				newData.append(StreamDownload.LINE_BREAK)
				newData.append(fileName.data(using: .utf8)!)
			} else {
				throw DownloadError("cannot interpret \(line) as a URL")
			}
		}
		
		let u: URL
		
		if let path = locationOfSegmentsM3u8 {
			u = target.appendingPathComponent(path)
		} else {
			u = indexUrl
		}
		
		if keepMaster {
			locationOfSegments = u.deletingLastPathComponent()
		}
		
		try writeData(newData, toFile: u)
	}

	private func downloadCompletionHandler(url: URL, targetFile: URL, data: Data?, response: URLResponse?, err: Error?) {
		if let error = err
		{
			destroy()
			downloader.theError = error
		}
		else if data == nil
		{
			destroy()
			downloader.theError = DownloadError("got no error and no data from \(url)")
		}
		else
		{
			do {
				let segment = data!
				let n = Int64(segment.count)

				internalSynchronizationQueue.sync {
					downloadedBytes += n
					events.append(Event(time: Date(), bytes: downloadedBytes))
				}
				try writeData(segment, toFile: targetFile)
			} catch {
				destroy()
				downloader.theError = error
			}
		}

		_ = internalSynchronizationQueue.sync {
			downloadTasks.removeValue(forKey: url)
		}

	}
	
	private func downloadSegments(base: URL) throws
	{
		state = .downloading

		if segmentUrl.count == 1 {
			let url: URL = segmentUrl.first!

			os_log("downloading video track %@", type: .debug, url.absoluteString)
			target = targetFile(url)
			download(url)
		} else {
			for tsUrl in segmentUrl
			{
				if state == .cancelling || state == .cancelled {
					break
				}
				try startDownloadTask(url: tsUrl, targetFile: targetFile(tsUrl))
			} // for
			
			while downloadTasks.count > 0 {
				Thread.sleep(forTimeInterval: 0.1)
			}

			state = .finished
		}
	}
	
	private func targetFile(_ url: URL) -> URL
	{
		let fileName = HtmlUtil.stripUrlParameters(url.lastPathComponent)
		
		return
			URL(fileURLWithPath: fileName, isDirectory: false, relativeTo: self.locationOfSegments ?? self.target)
	}

/// - throws: `Error`, `DownloadError`
  private func writeData(_ data: Data, toFile file: URL) throws
  {
	  let parent = file.deletingLastPathComponent()
	  let files = Downloader.FILES
	  var isDir: ObjCBool = true
	  let parentPath: String = parent.path

	  if !existingFolders.contains(parentPath) {
		  if files.fileExists(atPath: parentPath, isDirectory: &isDir) {
			  if !isDir.boolValue {
				  throw DownloadError("\(parentPath) exists, but is not a folder")
			  }
		  } else {
			  try files.createDirectory(at: parent, withIntermediateDirectories: true)
		  }
		  existingFolders.append(parentPath)
	  }
	  
	  let path: String = file.path
	  
	  if !files.createFile(atPath: path, contents: data)
	  {
		  throw DownloadError("could not create file \(path)")
	  }
	  
	  try files.setAttributes(Downloader.FILE_ATTR, ofItemAtPath: path)
  }
}
