//
//  DownloadsWindowController.swift
//  Filmbeschaffer
//
//  Created by Jakob on 20.5.18.
//  Copyright © 2018 Jakob. All rights reserved.
//

import Cocoa
import os.log
import AppUtilLib

class DownloadsWindowController:
	NSWindowController, NSTableViewDataSource, NSTableViewDelegate, NSUserInterfaceValidations
{
	//MARK: Public Static

	static func create() -> DownloadsWindowController
	{
		return DownloadsWindowController(windowNibName: "DownloadsWindow")
	}

	//MARK: Public Instance

	@IBOutlet var table: NSTableView!

	func add(download: Download){
		let k = downloads.count

		downloads.append(download)
		self.table.insertRows(at: IndexSet(integer: k))

		if download is FileDownload {
			download.start()
		}
	}

	func show(){
		guard window != nil else {
			NSSound.beep()
			return
		}

		window!.makeKeyAndOrderFront(self)

		for col in table.tableColumns {
			let title = col.title
			let id = col.identifier

			if title.range(of: "URL") != nil {
				urlCol = id
			} else if title.range(of: "File") != nil {
				fileCol = id
			} else if title.range(of: "Stat") != nil {
				statusCol = id
				variableColumns.insert(table.column(withIdentifier: id))
			} else if title.range(of: "Total") != nil {
				totalCol = id
				variableColumns.insert(table.column(withIdentifier: id))
			} else if title.range(of: "Downloaded") != nil {
				downloadedCol = id
				variableColumns.insert(table.column(withIdentifier: id))
			} else if title.range(of: "Speed") != nil {
				speedCol = id
				variableColumns.insert(table.column(withIdentifier: id))
			} else if title.range(of: "Time") != nil {
				timeCol = id
				variableColumns.insert(table.column(withIdentifier: id))
			}
		}

		table.doubleAction = #selector(openSelected)

		Timer.scheduledTimer(withTimeInterval: 1, repeats: true){
			t in self.refresh()
		}
	}

	//MARK: IBActions

	@IBAction
	func cancelDownloads(_ sender: AnyObject){
		for i in focussedRows {
			downloads[i].cancel()
		}
	}

	@IBAction
	func openVideo(_ sender: AnyObject){
		openSelected()
	}

	@IBAction
	func delete(_ sender: AnyObject){
		let set = focussedRows

		table.removeRows(at: set, withAnimation: DownloadsWindowController.ROW_ANIMATION)

		for i in set.sorted(by: >) // descending order, otherwise בלגן
		{
			downloads.remove(at: i).destroy()
		}
	}

	@IBAction
	func resume(_ sender: AnyObject){
		for i in focussedRows {
			let dl = downloads[i]

			if dl.isResumable {
				dl.start()
			}
		}
	}

	@IBAction
	func copyUrl(_ sender: AnyObject){
		if let dl = focussedRows.first {
			OsUtil.copyToClipboard(downloads[dl].url.description)
		}
	}

	//MARK: NSUserInterfaceValidations

	func validateUserInterfaceItem(_ item: NSValidatedUserInterfaceItem) -> Bool
	{
		let foc = focussedRows
		var active: Bool = !foc.isEmpty

		guard active else {
			return active
		}
		
		switch(item.action)
		{
			case #selector(cancelDownloads(_:)):
				active = areAll(foc){
					$0.isActive
				}

			case #selector(resume(_:)):
				active = areAll(foc){
					$0.isResumable
				}
	
			case #selector(openVideo(_:)):
				active = areAll(foc){
					$0.isFinished
				}

			case #selector(copyUrl(_:)): active = foc.count == 1
				
			case #selector(delete(_:)): break

			default: fatalError("unexpected action " + String(describing: item.action))
		}

		return active
	}

	//MARK: NSTableViewDataSource

	func numberOfRows(in tableView: NSTableView) -> Int
	{
		return downloads.count
	}

	//MARK: NSTableViewDelegate

	func tableView(_ tableView: NSTableView, viewFor tableColumn: NSTableColumn?, row: Int) -> NSView?
	{
		let colId = tableColumn!.identifier
		let cellView = tableView.makeView(withIdentifier: colId, owner:self) as! NSTableCellView
		let text = cellView.textField!
		let download = downloads[row]
		var uiText = ""

		switch colId
		{
			case urlCol:
				uiText = download.url.absoluteString

			case fileCol:
				uiText = download.target.path

			case statusCol:
				switch(download.status){
					case .preparing:
						uiText = "🍳"
					case .downloading:
						uiText = "🏃‍♀️"
					case .cancelled:
						uiText = "canceled"
					case .error:
						let err = download.errorMessage
						uiText = err == nil ? "⚡" : "⚡ " + err!
					case .cancelling:
						uiText = "canceling"
					case .waiting:
						uiText = "🧍‍♀️"
					default:
						uiText = "✅"
				}

			case totalCol:
				uiText = DownloadsWindowController.toMiB(download.totalSize)

			case downloadedCol:
				uiText = DownloadsWindowController.toMiB(download.downloaded)

			case speedCol:
				if download.status == .downloading {
					uiText = DownloadsWindowController.toIntString(download.speed / DownloadsWindowController.KI)
				}

			case timeCol:
				if download.status == .downloading {
					let speed = download.speed
					let totalSize: Int64 = download.totalSize
					
					if speed > 0 && totalSize != 0 {
						let s = Int((Double(totalSize - download.downloaded) / speed).rounded())
						var m = s / 60

						if m != 0 {
							uiText = "\(m)' "
						}

						m = s % 60

						if m != 0 || uiText.isEmpty {
							uiText += "\(m)\""
						}
					}
				}

			default:
				uiText = "unexpected column ID \"\(colId)\""
		}

		text.stringValue = uiText
		return cellView
	}


	//MARK: Private Static

	private static let MI10 = Double(1 << 20) / 10
	private static let KI = Double(1 << 10)
	private static let ROW_ANIMATION: NSTableView.AnimationOptions = [.effectFade, .slideUp]

	private static func toMiB(_ x: Int64) -> String
	{
		var y = toIntString(Double(x) / MI10)
		let n = y.removeLast()

		return (n == "0" ? y : y + "." + String(n))
	}

	private static func toIntString(_ x: Double) -> String
	{
		return String(Int(x.rounded()))
	}

	//MARK: Private Instance

	private var downloads = [Download]()
	private var urlCol: NSUserInterfaceItemIdentifier?
	private var fileCol: NSUserInterfaceItemIdentifier?
	private var statusCol: NSUserInterfaceItemIdentifier?
	private var totalCol: NSUserInterfaceItemIdentifier?
	private var downloadedCol: NSUserInterfaceItemIdentifier?
	private var speedCol: NSUserInterfaceItemIdentifier?
	private var timeCol: NSUserInterfaceItemIdentifier?
	private var variableColumns = IndexSet()
	
	private var focussedRows: IndexSet
	{
		let sel = table.selectedRowIndexes
		let clicked = table.clickedRow
		let res: IndexSet
		let haveClicked: Bool = clicked >= 0
		
		if !haveClicked || sel.contains(clicked){
			res = sel
		} else if haveClicked && sel.isEmpty{
			res = IndexSet(integer: clicked)
		} else {
			res = IndexSet()
		}
		return res
	}

	private func refresh(){
		let n = downloads.count
		var isStreamDownloading = false
		var waitingDownload: Download?

		for i in 0 ..< n {
			let dl = downloads[i]

			// If !isFinished we have to update the view even if the state is not changed.
			if !dl.isFinished || dl.isNewState {
				DispatchQueue.main.async {
					self.table.reloadData(forRowIndexes: IndexSet(integer: i), columnIndexes: self.variableColumns)
				}
			}
			if dl.status == .waiting {
				waitingDownload = dl
			}
			if dl.isActive && dl is StreamDownload {
				isStreamDownloading = true
			}
		}

		if !isStreamDownloading {
			if let sdl = waitingDownload {
				sdl.start()
			}
		}
	}

	@objc
	private func openSelected(){
	  for i in focussedRows {
		  downloads[i].openVideo()
	  }
  }

	private func areAll(_ selected: IndexSet, crit: (Download) -> Bool) -> Bool
	{
		var active = true

		for i in selected {
			if !crit(downloads[i]) {
				active = false
				break
			}
		}
		return active
	}
}
